# Importing FedBizOpps data into FBOpen using the fbo.gov loader

## Limitations
* The fbo.gov loader currently loads only **COMBINE** and **PRESOL** opportunity notices, and only loads notices for which the due date has not yet passed.
* **The loader does not yet handle MOD notices.** This is high priority on the open issues list.
* The attachment crawler/loader is still very primitive. It works for attachments uploaded to fbo.gov or to which the listing's synopsis links directly. Attachments that are more than one hop removed from the listing page are *not* retrieved, but we hope to improve the crawler to do this.

## To do
* load MOD notices, updating existing data
* load AWARD and other notices types
* survey FBO listing archetypes to determine the algorithms for retrieving the relevant attachments for various types of notices, and implement that logic in the crawler/loader
* better failure handling for the metadata loader, attachment crawler, and attachment loader
* script to automatically run the nightly update

## To install the loader
* Install Solr. See the README in `/solr-files` .
* In the loaders/fbo.gov/ directory:
	* `npm install`
	* `mkdir workfiles nightly-downloads fbo-attachments tmp`

	**Note:** you can configure the fbo.gov loader to use a different directory for fbo attachments by editing `fbo-loader-config.js`.

## To import FedBizOpps data
After install, first load a full set of data using the one-time/weekly loader. Then you can simply run nightly updates.

### 1. One-time/Weekly full load
1. **`get-fbo-xml.sh`**

	Downloads the full, weekly FBO XML file from ftp://ftp.fbo.gov.
	
	Optional arguments:
	
	`$1` = where to save the downloaded file; default = *./workfiles/FBOFullXML.xml*

2. **`fbo-solrize-big.sh`**

	From the fbo.gov weekly XML data dump, creates a Solr-ingestible XML file.
	
	Optional arguments:

	`$1` = filepath/name of a weekly FBO XML "dump" file; default = *./workfiles/FBOFullXML.xml*

 	`$2` = output filepath/name for the "solrized" xml (which will be imported into Solr); default = *./workfiles/listings-solrized.xml*

 	`$3` = output filepath/name of the list of links to the FBO listings that were solrized; default = *./workfiles/listings-links.txt*

 	`$4` = directory into which attachments should be downloaded; default = *./fbo-attachments*


3. **`load-solrized.sh`**

	Loads the Solr-ingestible XML file into Solr.
	
	Optional arguments:

	`$1` output xml from step 2; default = *./workfiles/listings-solrized.xml*

	**Note:** in `load-solrized.sh`, make sure `postjarfile` and `solrlogdir` point to your Solr installation. [For our to-do list: automate this, or make it separately and more obviously configurable.]


4. **`process-listing-links.sh < workfiles/listings-links.txt`** *(or alternate filename of list of links generated by step 2)*

	Downloads and loads into Solr the relevant attachments from each URL in the input stream. Assumes those URLs are fbo.gov listing pages.
	
	**NOTE 1:** This process is currently very slow, as it's completely synchronous. This is intentional; trying to download too many files at once, and trying to load too many into Solr at once, led to problems. Better would be (1) a throttled async version that handles several file downloads simultaneously, and (2) Do the Solr ingestion in a completely separate step.
	
	**NOTE 2:** It's also very buggy. There are many files that shouldn't be downloaded but are, and there are many files that are behind multiple clicks and/or unreachable without login. The algorithm for doing these downloads will have to improve significantly.
	
### Nightly updates
All at once: **`fbo-nightly.sh [YYYYMMDD]`** (defaults to yesterday)

Or, if you want to do it in steps:

* To download the nightly data file: `node nightly-fbo-parser.js -o [-d YYYYMMDD]` (defaults to yesterday)
* To load the nightly data into FBOpen: `node nightly-fbo-parser.js [-d YYYYMMDD]` (defaults to yesterday)
* To collect and load listings' attachments into FBOpen: `process-listing-links.sh < links-YYYYMMDD.txt`

The parser in the nightly code -- that is, the hard part -- is borrowed wholesale from our predecessor [Adam Becker's](https://github.com/adamjacobbecker/) [fbo-parser](https://github.com/presidential-innovation-fellows/fbo-parser). Thanks Adam!
